@@ -79,7 +79,6 @@ static int current_function_is_inlinable;
 
 static rtx guard_area, _guard;
 static rtx function_first_insn, prologue_insert_point;
-static rtx debuginsn;
 
 /*  */
 static HOST_WIDE_INT sweep_frame_offset;
@@ -906,7 +905,7 @@ arrange_var_order (block)
 
 
 static void
-copy_args_for_protection (void)
+copy_args_for_protection ()
 {
   tree parms = DECL_ARGUMENTS (current_function_decl);
   rtx temp_rtx;
@@ -1250,6 +1249,7 @@ sweep_string_use_of_insns (insn, sweep_offset, sweep_size)
       {
 	has_virtual_reg = FALSE;
 	sweep_string_in_operand (insn, &PATTERN (insn), sweep_offset, sweep_size);
+	sweep_string_in_operand (insn, &REG_NOTES (insn), sweep_offset, sweep_size);
       }
 }
 
@@ -1522,6 +1522,29 @@ change_arg_use_in_operand (x, orig, new, size)
 	}
       break;
 
+    case SET:
+      /* Handle special case of "set (REG or MEM) (incoming_args)".
+	 It means that the the address of the 1st argument is stored. */
+      if (GET_CODE (orig) == MEM
+	  && XEXP (x, 1) == virtual_incoming_args_rtx)
+	{
+	  offset = 0;
+
+	  /* the operand related to the sweep variable */
+	  if (AUTO_OFFSET(XEXP (orig, 0)) <= offset &&
+	      offset < AUTO_OFFSET(XEXP (orig, 0)) + size) {
+
+	    offset = AUTO_OFFSET(XEXP (*new, 0))
+	      + (offset - AUTO_OFFSET(XEXP (orig, 0)));
+
+	    XEXP (x, 1) = plus_constant (virtual_stack_vars_rtx, offset);
+	    XEXP (x, 1)->used = 1;
+
+	  return;
+	  }
+	}
+      break;
+
     case CALL_PLACEHOLDER:
       change_arg_use_of_insns_2 (XEXP (x, 0), orig, new, size); if (flag_caui_exit) return;
       change_arg_use_of_insns_2 (XEXP (x, 1), orig, new, size); if (flag_caui_exit) return;
@@ -1975,8 +1998,14 @@ push_frame_in_args (parms, push_size, boundary)
 	    /* the operand related to the sweep variable */
 	    if (AUTO_BASEPTR (XEXP (home, 0)) == frame_pointer_rtx)
 	      {
-		offset += push_size;
-		XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode, offset);
+		if (XEXP (home, 0) == frame_pointer_rtx)
+		  XEXP (home, 0) = plus_constant (frame_pointer_rtx,
+						  push_size);
+		else {
+		  offset += push_size;
+		  XEXP (XEXP (home, 0), 1) = gen_rtx_CONST_INT (VOIDmode,
+								offset);
+		}
 
 		/* mark */
 		XEXP (home, 0)->used = 1;
@@ -2001,23 +2030,28 @@ push_frame_of_insns (insn, push_size, boundary)
     if (GET_CODE (insn) == INSN || GET_CODE (insn) == JUMP_INSN
 	|| GET_CODE (insn) == CALL_INSN)
       {
-	insn_pushed = FALSE; debuginsn = insn;
+	insn_pushed = FALSE;
 	push_frame_in_operand (insn, PATTERN (insn), push_size, boundary);
 
-	if (insn_pushed
-	    /* skip if it is called from purge_address_of */
-	    && cse_not_expected)
+	if (insn_pushed)
 	  {
-	    rtx after = insn;
+	    rtx trial = insn;
+	    rtx before = PREV_INSN (trial);
+	    rtx after = NEXT_INSN (trial);
+	    int has_barrier = 0;
+	    rtx tem;
 	    rtx seq = split_insns (PATTERN (insn), insn);
 
+	    /* If we are splitting a JUMP_INSN, it might be followed by a 
+	       BARRIER. We may need to handle this specially.  */
+	    if (after && GET_CODE (after) == BARRIER)
+	      {
+		has_barrier = 1;
+		after = NEXT_INSN (after);
+	      }
+
 	    if (seq && GET_CODE (seq) == SEQUENCE)
 	      {
-		register int i;
-
-		/* replace the pattern of the insn */
-		PATTERN (insn) = PATTERN (XVECEXP (seq, 0, 0));
-
 		if (XVECLEN (seq, 0) == 2)
 		  {
 		    rtx pattern = PATTERN (XVECEXP (seq, 0, 1));
@@ -2030,21 +2064,26 @@ push_frame_of_insns (insn, push_size, boundary)
 		      {
 			rtx offset = XEXP (XEXP (pattern, 1), 1);
 			fp_equiv[REGNO (XEXP (pattern, 0))] = INTVAL (offset);
+
+			/* replace the pattern of the insn */
+			add_insn_after (XVECEXP (seq, 0, 0), before);
+			delete_insn (trial);
 			goto next;
 		      }
 		  }
-		
-		for (i = 1; i < XVECLEN (seq, 0); i++)
-		  {
-		    rtx insn = XVECEXP (seq, 0, i);
-		    add_insn_after (insn, after);
-		    after = insn;
-		  }
+
+		/* excerpt from emit-rtl.c: L3320 */
+		tem = emit_insn_after (seq, trial);
+
+		delete_related_insns (trial);
+		if (has_barrier)
+		  emit_barrier_after (tem);
 
 		/* Recursively call try_split for each new insn created */
-	        insn = NEXT_INSN (insn);
-		for (i = 1; i < XVECLEN (seq, 0); i++, insn = NEXT_INSN (insn))
-		  insn = try_split (PATTERN (insn), insn, 1);
+		for (tem = NEXT_INSN (before); tem != after;
+		     tem = NEXT_INSN (tem))
+		  if (! INSN_DELETED_P (tem) && INSN_P (tem))
+		    tem = try_split (PATTERN (tem), tem, 1);
 	      }
 	  }
 
